"""
This is the main command-line interface for performing optimisation on a CNN
model. Example usage is shown below:

```shell
usage: python -m samo [-h] --model PATH -b {fpgaconvnet,finn,hls4ml} -p PATH
        -o PATH [--optimiser {brute,annealing,init,rule}]
        [--objective {throughput,latency}] [--enable_reconf {true,false}]
        [--seed N] [--batch-size N]

optional arguments:
  -h, --help            show this help message and exit
  -m PATH, --model PATH
                        path to the CNN model that you wish to
                        optimise (.keras, .onnx)
  -b {fpgaconvnet,finn,hls4ml}, --backend {fpgaconvnet,finn,hls4ml}
                        target backend for accelerating the model
  -p PATH, --platform PATH
                        hardware platform details (.json)
  -o PATH, --output-path PATH
                        output path for the optimised model (.json, .onnx)
  --optimiser {brute,annealing,init,rule}
                        optimiser to use
  --objective {throughput,latency}
                        Optimiser objective
  --enable_reconf {true,false}
                        multiple partitions
  --seed N              Seed for the optimiser run
  --batch-size N        batch size
```
"""

import argparse
import importlib
import json
import copy
import time
import random
import sys

import numpy as np

from samo.optimiser.annealing import SimulatedAnnealing
from samo.optimiser.rule import RuleBased
from samo.optimiser.brute import BruteForce

def main(args):
    # parse arguments
    parser = argparse.ArgumentParser(description="SAME CNN optimiser")
    parser.add_argument("-m", "--model", metavar="PATH", required=True,
            help="path to the CNN model that you wish to optimise (.keras, .onnx)")
    parser.add_argument("-b", "--backend", choices=["fpgaconvnet", "finn", "hls4ml"], required=True,
            help="target backend for accelerating the model")
    parser.add_argument("-p", "--platform", metavar="PATH", required=True,
            help="hardware platform details (.json)")
    parser.add_argument("-o", "--output-path", metavar="PATH", required=True,
            help="output path for the optimised model (.json, .onnx)")
    parser.add_argument("--optimiser", choices=["brute", "annealing", "init", "rule"], required=False, default="annealing",
            help="optimiser to use")
    parser.add_argument('--objective', choices=['throughput','latency'], required=False, default="latency", help='Optimiser objective')
    parser.add_argument("--enable_reconf", choices=["true", "false"], required=False, default="true", help="multiple partitions")
    parser.add_argument('--seed', metavar='N', type=int, default=random.randint(0,2**32-1),
        help='Seed for the optimiser run')
    parser.add_argument('--batch-size', metavar='N', type=int, default=256,
        help='batch size')

    args = parser.parse_args(args)

    # get the batch size
    batch_size = args.batch_size
    if args.objective == "latency":
        batch_size = 1

    # print the run setting
    print("#### ( run settings ) ####")
    print(f" * model    : {args.model}")
    print(f" * backend  : {args.backend}")
    print(f" * platform : {args.platform}")
    print(f" * optimiser : {args.optimiser}")
    print(f" * objective : {args.objective}")
    print(f" * batch size : {args.batch_size}")
    print(f" * enable_reconf : {args.enable_reconf}")
    print(f" * seed : {args.seed}")
    print(f" * output_path : {args.output_path}")
    print("##########################")

    # parse the platform
    with open(args.platform, "r") as f:
        platform = json.load(f)

    # get the backend parser and exporter
    parser = importlib.import_module(f"samo.backend.{args.backend}.parser")
    exporter = importlib.import_module(f"samo.backend.{args.backend}.export")

    # parse the network
    graph = parser.parse(args.model, platform, batch_size)

    graph.enable_reconf = {"true":True, "false":False}[args.enable_reconf]
    graph.objective = args.objective

    random.seed(args.seed)
    np.random.seed(args.seed)

    # init
    for partition in graph.partitions:
        partition.reset()

    # create an optimiser instance for the network
    if args.optimiser == "annealing":
        opt = SimulatedAnnealing(graph)
    elif args.optimiser == "rule":
        opt = RuleBased(graph)
    elif args.optimiser == "brute":
        opt = BruteForce(graph)
    elif args.optimiser == "init":
        graph.summary()
        exporter.export(graph, args.model, args.output_path)
        return
    else:
        raise NameError

    opt.start_time = time.time()

    # split up the network completely
    can_split = args.optimiser != "brute"
    while can_split:
        can_split = False
        for i in range(len(opt.network.partitions)):
            valid_splits = opt.network.valid_splits(i)
            network_copy = copy.deepcopy(opt.network)
            if valid_splits:
                can_split = True
                prev = opt.network.check_constraints()
                opt.network.split(i, valid_splits[0])
                if prev and not opt.network.check_constraints():
                    can_split = False
                    opt.network = network_copy

    # validate generated design
    assert opt.network.check_constraints(), "Intial design infeasible!"

    # run the optimiser
    opt.optimise()

    # validate generated design
    assert opt.network.check_constraints(), "Optimised design infeasible!"

    # print a summary of the run
    opt.network.summary()

    # export the design
    exporter.export(opt.network, args.model, args.output_path)

if __name__ == "__main__":
    main(sys.argv[1:])



import itertools
import copy
from dataclasses import dataclass
import numpy as np
from tqdm import tqdm

from samo.model.network import Network

@dataclass
class BruteForce:
    network: Network

    def update(self, config):
        assert len(self.network.partitions) == 1

        layers = []
        for index, layer in enumerate(self.network.partitions[0]):
            self.network.partitions[0].nodes[layer]["hw"].channel_in_folding    = config[index][0]
            self.network.partitions[0].nodes[layer]["hw"].channel_out_folding   = config[index][1]
            self.network.partitions[0].nodes[layer]["hw"].kernel_folding        = config[index][2]
            self.network.partitions[0].nodes[layer]["hw"].update(hw_update=True)
            layers.append(layer)


        index = 0
        while index < len(config):
            if config[index][3]:
                for partition_index, partition in enumerate(self.network.partitions):
                    if layers[index] in partition:
                        for next_node in partition.successors(layers[index]):
                            self.network.split(partition_index, (layers[index], next_node))
            index += 1


    def optimise(self):
        assert len(self.network.partitions) == 1

        # get all the configurations
        size = 1
        configurations = []
        for layer in self.network.partitions[0]:

            layer_split = [False]
            for split_pair in self.network.valid_splits(0):
                if split_pair[0] == layer:
                    layer_split = [False, True]
                    break

            layer_configurations = list(itertools.product(
                self.network.partitions[0].nodes[layer]["hw"].valid_channel_in_folding,
                self.network.partitions[0].nodes[layer]["hw"].valid_channel_out_folding,
                self.network.partitions[0].nodes[layer]["hw"].valid_kernel_folding,
                layer_split
                ))

            configurations.append(layer_configurations)
            size *= len(layer_configurations)
        configurations = itertools.product(*configurations)

        print(f"full configuration space size : {size}")

        for i, layer in enumerate(tqdm(self.network, desc="collecting inter folding matching configurations")):
            if self.network.nodes[layer]["hw"].constraints["matching_inter_folding"] and self.network.out_degree(layer) > 0:
                configurations = filter(lambda x,i=i: x[i][1] == x[i+1][0], configurations)
            if self.network.nodes[layer]["hw"].constraints["divisible_inter_folding"] and self.network.out_degree(layer) > 0:
                configurations = filter(lambda x,i=i: max(x[i][1], x[i+1][0]) % min(x[i][1], x[i+1][0]) == 0, configurations)

        size = 0
        for _ in tqdm(copy.deepcopy(configurations), desc="counting space size"):
            size += 1

        network_init = copy.deepcopy(self.network)

        # track all valid networks
        valid_configs = {}

        # iterate over all the configurations
        for config in tqdm(configurations, desc="evaluating configurations"):
            # update the network
            self.network = copy.deepcopy(network_init)
            self.update(config)
            # evaluate the cost
            cost  = self.network.eval_cost()
            # if network is within constraints, log the network and it's cost
            if self.network.check_constraints():
                valid_configs[config] = cost

        # find the network with the lowest cost
        best_config = min(valid_configs, key=valid_configs.get_cost)

        # update with the best configuration
        self.network = copy.deepcopy(network_init)
        self.update(best_config)



from fpgaconvnet_optimiser.models.layers import Layer, ConvolutionLayer, InnerProductLayer
from samo.model import Node

class FPGAConvNetWrapper(Node):

    def __init__(self, layer: Layer, batch_size=1):

        # store the batch size
        self.batch_size = batch_size

        # store the fpgaconvnet layer
        self.layer = layer

        # get the size in and out
        self.size_in = layer.rows_in()*layer.cols_in()*layer.channels_in()
        self.size_out = layer.rows_out()*layer.cols_out()*layer.channels_out()

        # get the channels in and out for the layer
        self.channels_in    = layer.channels_in()
        self.channels_out   = layer.channels_out()

        # add the kernel size if it's a convolution layer
        if type(layer) == ConvolutionLayer:
            self.kernel_size = layer.kernel_size[0]

        # set the matching folding constraint
        self.constraints = { "matching_intra_folding" : type(layer) not in [ConvolutionLayer, InnerProductLayer],
                             "matching_inter_folding" : False,
                             "divisible_inter_folding" : False}

    def update(self, hw_update=False):
        self.layer.coarse_in = self.channel_in_folding
        self.layer.coarse_out = self.channel_out_folding
        if hasattr(self.layer, "fine"):
            self.layer.fine = self.kernel_folding
        if hw_update:
            self.layer.update()

    def latency(self):
        return self.layer.latency()*self.batch_size

    def resource(self):
        return self.layer.resource()




"""
This part of the repo contains high-level models of Streaming Architecture
building blocks and their connectivity. In particular, there are three levels
of abstraction that are used:

- `samo.model.node.Node`: fundamental building blocks of the accelerator
- `samo.model.partition.Partition`: subgraph of nodes which constitute an FPGA
configuration
- `samo.model.network.Network`: a collection of partitions which create the
functionality of the whole CNN model
"""

from .node import Node
from .network import Network
from .partition import Partition




from fpgaconvnet_optimiser.models.network.Network import Network
from fpgaconvnet_optimiser.models.partition.Partition import Partition
import fpgaconvnet_optimiser.tools.graphs as graphs

def export(network, model_path, output_path):

    # parse the model as a Network
    fpgaconvnet_net = Network("same", model_path)

    # set partitions to being empty
    fpgaconvnet_net.partitions = []

    # iterate over partitions
    for partition_index in range(len(network.partitions)):
        # get all nodes in partition
        nodes = list(network.partitions[partition_index].nodes)
        # get a subgraph of the network and append it to partitions
        fpgaconvnet_net.partitions.append(Partition(
            fpgaconvnet_net.graph.subgraph(nodes).copy()))
        # update nodes in partition
        for node in nodes:
            fpgaconvnet_net.partitions[-1].graph.nodes[node]["hw"] = \
                    network.partitions[partition_index].nodes[node]["hw"].layer

    # update the network
    fpgaconvnet_net.update_partitions()

    # generate the output configuration
    fpgaconvnet_net.save_all_partitions(output_path)



import csv
import math
import copy
import random
from dataclasses import dataclass, field
import numpy as np
import time
from tqdm import tqdm

from samo.model import Network

@dataclass
class SimulatedAnnealing:
    network: Network
    T: float = 10.0
    k: float = 100.0
    T_min: float = 0.001
    cool: float = 0.995
    iterations: int = 10
    valid_variables: list = field(default_factory=lambda: ["channel_in_folding", "channel_out_folding", "kernel_folding"])

    def update(self):
        for partition in self.network.partitions:
            for index, layer in enumerate(partition):
                partition.nodes[layer]["hw"].update(hw_update=True)

    def random_transformation(self):
        # choose to do a partitioning transform of change a variable
        transform = np.random.choice(["partition", "variable"], p=[0.1,0.9])
        # pick a random partition
        partition = random.choice(self.network.partitions)
        # get partition index
        partition_index = self.network.partitions.index(partition)

        # pick a random transform
        if transform == "partition":
            transform_type = random.choice(["split", "merge"])
            if transform_type == "split":
                # choose random split nodes
                valid_splits = self.network.valid_splits(partition_index)
                if valid_splits:
                    nodes = random.choice(valid_splits)
                    # apply split
                    self.network.split(partition_index, nodes)
            elif transform_type == "merge":
                # choose random merge
                valid_merges = self.network.valid_merges()
                if valid_merges:
                    merge = random.choice(valid_merges)
                    self.network.merge(merge)
        else:
            # pick a random layer
            layer = random.choice(list(partition.nodes()))
            node_hw = partition.nodes[layer]["hw"]
            # pick a random variable
            variable = random.choices(self.valid_variables)[0]
            # apply a random value to that variable (within constraints)
            if variable == "channel_in_folding":
                folding = random.choices(node_hw.valid_channel_in_folding)[0]
                node_hw.channel_in_folding = folding
                partition.folding_match(layer, folding, "io")
            elif variable == "channel_out_folding":
                folding = random.choices(node_hw.valid_channel_out_folding)[0]
                node_hw.channel_out_folding = folding
                partition.folding_match(layer, folding, "io")
            elif variable == "kernel_folding":
                node_hw.kernel_folding = random.choices(node_hw.valid_kernel_folding)[0]

    def optimise(self):

        def generator():
            while self.T_min < self.T:
                yield

        log = []

        # keep iterating until we meet the minimum temperature
        pbar = tqdm(generator())
        for _ in pbar:

            # update the description
            pbar.set_description(desc=f"simulated annealing iterations (T={self.T:.3f})")

            # get the throughput of the current network state
            cost = self.network.eval_cost()

            # keep a copy of the current network state
            network_copy = copy.deepcopy(self.network)

            # perform a number of permutations of this network
            for _ in range(self.iterations):
                self.random_transformation()

            # update the network
            self.update()

            # for partition in self.network.partitions:
            #     print(partition.eval_latency(), partition.eval_resource())

            # check the network is within platform resource constraints
            if not self.network.check_constraints():
                self.network = network_copy
                continue


            # log the current resources and cost
            new_cost = self.network.eval_cost()
            # new_resource = self.network.eval_resource()
            chosen = True

            # perform the annealing descision
            if math.exp(min(0,(latency - new_latency)/(self.k*self.T))) < random.uniform(0,1):
                self.network = network_copy
                chosen = False

            # reduce temperature
            self.T *= self.cool

            # # update the log
            if chosen:
                log += [[
                        time.time()-self.start_time,
                        new_cost,
                        #new_resource["BRAM"],
                        #new_resource["DSP"],
                        #new_resource["LUT"],
                        #new_resource["FF"],
                ]]

        # write log to a file
        with open("outputs/log.csv", "w") as f:
            writer = csv.writer(f)
            [ writer.writerow(row) for row in log ]



from dataclasses import dataclass, field
from functools import reduce
from tabulate import tabulate
import networkx as nx
import json

class Partition(nx.DiGraph):
    freq: float = 100.0
    wordlength: int = 16
    platform: dict = {
            "resources": {
                "BRAM": 0,
                "DSP": 0,
                "FF": 0,
                "LUT": 0
            },
            "bandwidth": 0.0,
            "reconf_time": 0.0
        }
    constraints: dict = {"resource" : True, "inter_layer_matching" : True}

    @property
    def input_node(self):
        return [ edge for edge, deg in self.in_degree() if not deg ][0]

    @property
    def output_node(self):
        return [ edge for edge, deg in self.out_degree() if not deg ][0]

    def eval_latency(self):
        return max([ self.nodes[layer]["hw"].latency() for layer in self.nodes])

    def eval_throughput_in(self):
        return self.nodes[self.input_node]["hw"].size_in/self.eval_latency() * \
                self.freq * self.wordlength

    def eval_throughput_out(self):
        return self.nodes[self.output_node]["hw"].size_out/self.eval_latency() * \
                self.freq * self.wordlength

    def eval_resource(self):
        return {
            "LUT"   : sum([ self.nodes[layer]["hw"].resource()["LUT"]  for layer in self.nodes]),
            "DSP"   : sum([ self.nodes[layer]["hw"].resource()["DSP"]  for layer in self.nodes]),
            "BRAM"  : sum([ self.nodes[layer]["hw"].resource()["BRAM"] for layer in self.nodes]),
            "FF"    : sum([ self.nodes[layer]["hw"].resource()["FF"]   for layer in self.nodes])
        }

    def check_resource_constraints(self):
        # get the current resource usage
        resource = self.eval_resource()
        # check within the platform constraints
        rsc_constraints = []
        rsc_constraints += [resource["BRAM"] <= self.platform["resources"]["BRAM"]]
        rsc_constraints += [resource["FF"]   <= self.platform["resources"]["FF"]]
        rsc_constraints += [resource["LUT"]  <= self.platform["resources"]["LUT"]]
        rsc_constraints += [resource["DSP"]  <= self.platform["resources"]["DSP"]]
        # if network is within constraints, return true
        return reduce(lambda a, b: a and b, rsc_constraints)

    def check_memory_bandwdith_constraint(self):
        return (self.eval_throughput_in() + self.eval_throughput_out()) < 1000*float(self.platform["bandwidth"])

    def check_matching_inter_folding(self, node, next_node):
        return self.nodes[node]["hw"].channel_out_folding == self.nodes[next_node]["hw"].channel_in_folding

    def check_constraints(self):
        # check all the constraints (if they are required)
        for node in self.nodes:
            # print(node, self.nodes[node]["hw"].check_constraints())
            if not self.nodes[node]["hw"].check_constraints():
                return False

            if self.nodes[node]["hw"].constraints["matching_inter_folding"] and self.out_degree(node) > 0:
                next_node = list(self.successors(node))[0]
                assert self.check_matching_inter_folding(node, next_node)

        # check resource constraints
        resource_check = self.check_resource_constraints() if self.constraints["resource"] else True
        bandwidth_check = self.check_memory_bandwdith_constraint()
        # return resource_check and bandwidth_check
        return resource_check and bandwidth_check

    def avg_rsc_util(self):
        resource = self.eval_resource()
        avg_rsc_utli = 0.25 * (resource["BRAM"] / self.platform["resources"]["BRAM"]) \
                        + 0.25 * (resource["DSP"] / self.platform["resources"]["DSP"]) \
                        + 0.25 * (resource["LUT"] / self.platform["resources"]["LUT"]) \
                        + 0.25 * (resource["FF"] / self.platform["resources"]["FF"])
        return avg_rsc_utli

    def summary(self):
        # get a summary for the whole network
        latency = self.eval_latency()
        resources = self.eval_resource()
        network_summary = tabulate([[
            int(latency),
            f"{resources['DSP']} / {self.platform['resources']['DSP']}",
            f"{resources['BRAM']} / {self.platform['resources']['BRAM']}",
            f"{resources['LUT']} / {self.platform['resources']['LUT']}",
            f"{resources['FF']} / {self.platform['resources']['FF']}"
        ]], headers=["Latency (cycles)", "DSP", "BRAM", "LUT", "FF"])
        # get a summary for each layer
        layer_summary = []
        for node in self.nodes:
            layer = self.nodes[node]["hw"]
            layer_summary.append(
                [ node, int(layer.latency()), layer.resource()["DSP"],
                    layer.resource()["BRAM"], layer.resource()["LUT"], layer.resource()["FF"]] )
        layer_summary = tabulate(layer_summary, headers=["Layer", "Latency (cycles)", "DSP", "BRAM", "LUT", "FF"])
        bandwidth_summary = tabulate([[
            self.eval_throughput_in(),
            self.eval_throughput_out(),
            f"{self.eval_throughput_in()+self.eval_throughput_out()}/{1000*self.platform['bandwidth']}",
        ]], headers=["in (Mbps)", "out (Mbps)", "total (Mbps)"])

        # print the summary
        print(bandwidth_summary)
        print("\n")
        print(layer_summary)
        print("\n")
        print(network_summary)
        print("\n\n")

    def folding_match(self, node, folding, direction):
        node_hw = self.nodes[node]["hw"]

        if node_hw.constraints["matching_intra_folding"]:
            node_hw.channel_in_folding = folding
            node_hw.channel_out_folding = folding

        channel_in_folding = node_hw.channel_in_folding
        channel_out_folding = node_hw.channel_out_folding

        if self.in_degree(node) > 0 and "i" in direction:
            for prev_node in self.predecessors(node):
                prev_channel_out_folding = self.nodes[prev_node]["hw"].channel_out_folding
                if self.nodes[prev_node]["hw"].constraints["matching_inter_folding"] or \
                self.nodes[prev_node]["hw"].constraints["divisible_inter_folding"] and max(channel_in_folding, prev_channel_out_folding) % min(channel_in_folding, prev_channel_out_folding) != 0:
                    self.nodes[prev_node]["hw"].channel_out_folding = channel_in_folding
                    if "wi" in direction:
                        # weak propogate, only once
                        self.folding_match(prev_node, channel_in_folding, "")
                    elif self.out_degree(prev_node) > 1:
                        # branch, propogate back
                        self.folding_match(prev_node, channel_in_folding, "iwo")
                    else:
                        self.folding_match(prev_node, channel_in_folding, "i")

        if self.out_degree(node) > 0 and "o" in direction:
            for next_node in self.successors(node):
                next_channel_in_folding = self.nodes[next_node]["hw"].channel_in_folding
                if node_hw.constraints["matching_inter_folding"] or \
                node_hw.constraints["divisible_inter_folding"] and max(channel_out_folding, next_channel_in_folding) % min(channel_out_folding, next_channel_in_folding) != 0:
                    self.nodes[next_node]["hw"].channel_in_folding = channel_out_folding
                    if "wo" in direction:
                        # weak propogate, only once
                        self.folding_match(next_node, channel_out_folding, "")
                    elif self.in_degree(next_node) > 1:
                        # branch, propogate back
                        self.folding_match(next_node, channel_out_folding, "wio")
                    else:
                        self.folding_match(next_node, channel_out_folding, "o")

    def save_config(self, config_path):
        network_config = {}
        for node in self.nodes:
            node_hw = self.nodes[node]["hw"]

            node_config = {}
            node_config["channel_in_folding"] = node_hw.channel_in_folding
            node_config["channel_out_folding"] = node_hw.channel_out_folding
            node_config["kernel_folding"] = node_hw.kernel_folding

            network_config[node] = node_config

        with open(config_path,"w") as f:
            json.dump(network_config,f,indent=2)

    def load_config(self, config_path):
        with open(config_path,"r") as f:
            network_config = json.load(f)

        for node in self.nodes:
            node_hw = self.nodes[node]["hw"]
            node_config = network_config[node]

            node_hw.channel_in_folding = node_config["channel_in_folding"]
            node_hw.channel_out_folding = node_config["channel_out_folding"]
            node_hw.kernel_folding = node_config["kernel_folding"]

        #if not self.check_constraints():
        #    print("Loading a config that may violate constraints")

        for node in self.nodes:
            self.nodes[node]["hw"].update(hw_update=True)

    def reset(self):
        for node in self.nodes:
            self.nodes[node]["hw"].reset()



import math
from functools import reduce
from dataclasses import dataclass, field
from typing import List

def get_factors(n):
    """
    helper function for returning all the factors of `n`
    """
    return list(set(reduce(list.__add__,
        ([i, n//i] for i in range(1, int(n**0.5) + 1) if n % i == 0))))

@dataclass
class Node:
    """
    A generalisation of hardware nodes in Streaming Architectures. This class
    represents the Node described in the paper. Hardware in the backend's
    parsed network representation is mapped to this class.
    """

    size_in: int
    """
    number of words from incoming feature-map into the node
    """
    size_out: int
    """
    number of words from outgoing feature-map from the node
    """
    channels_in: int
    """
    channel dimension of the incoming feature-map
    """
    channels_out: int
    """
    channel dimension of the outgoing feature-map
    """
    kernel_size: int = 1
    """
    kernel dimension of the node. Used to determine the valid kernel folding
    factors. The default is 1.
    """
    channel_in_folding: int = 1
    """
    channel in folding variable. Defines the level of parallelism across the
    incoming feature-map's channel dimension.
    The default is 1.
    """
    channel_out_folding: int = 1
    """
    channel out folding variable. Defines the level of parallelism across the
    outgoing feature-map's channel dimension.
    The default is 1.
    """
    kernel_folding: int = 1
    """
    kernel folding variable. This is parallelism across internal operations,
    defined by kernel size and valid kernel folding.
    The default is 1.
    """
    constraints: dict = field(default_factory=lambda: {"matching_intra_folding": True, "matching_inter_folding": False, "divisible_inter_folding": True})
    """
    dictionary of constraints on the node. Each key is a description of the
    constraint, and the value is a Boolean of whether it is to be applied or
    not.
    The defaults are:
    ```
    {
        "matching_intra_folding": True,
        "matching_inter_folding": False,
        "divisible_inter_folding": True
    }
    ```
    """

    def __hash__(self):
        return hash(repr(self))

    @property
    def valid_channel_in_folding(self) -> List[int]:
        return sorted(get_factors(self.channels_in))

    @property
    def valid_channel_out_folding(self):
        return sorted(get_factors(self.channels_out))

    @property
    def valid_kernel_folding(self):
        return sorted(get_factors(self.kernel_size*self.kernel_size))

    def check_matching_intra_folding(self):
        assert self.channel_in_folding == self.channel_out_folding
        return True

    def check_constraints(self):
        # check all the constraints
        constraints = []
        constraints += [self.check_matching_intra_folding if self.constraints["matching_intra_folding"] else True]
        constraints += [self.channel_in_folding in self.valid_channel_in_folding]
        constraints += [self.channel_out_folding in self.valid_channel_out_folding]
        constraints += [self.kernel_folding in self.valid_kernel_folding]
        # ensure all constraints are held
        return reduce(lambda a, b: a and b, constraints)

    def update(self, hw_update=False):
        pass

    def latency(self):
        return 1

    def resource(self):
        return {
             "LUT" : 0,
             "DSP" : 0,
             "BRAM" : 0,
             "FF" : 0
        }

    def reset(self):
        self.channel_in_folding = 1
        self.channel_out_folding = 1
        self.kernel_folding = 1

        self.update(hw_update=True)




import copy

import fpgaconvnet_optimiser.tools.parser as parser

from samo.model import Network
from samo.model import Partition

from .node import FPGAConvNetWrapper

def parse(filepath, platform, batch_size):

    # parse the network
    _, graph = parser.parse_net(filepath)

    # convert into the node wrappers
    reference = Partition()
    reference.platform = platform
    prev_node = None
    for node in graph.nodes:
        new_node = FPGAConvNetWrapper(graph.nodes[node]["hw"], batch_size=batch_size)
        reference.add_node(node, hw=new_node)
        # add edge to graph
        if prev_node != None:
            reference.add_edge(prev_node, node)
        prev_node = node

    # create network from reference design
    network = Network(reference)
    network.batch_size = batch_size

    # return the wrapped network
    return network



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="samo",
    version="0.2.0",
    author="Alex Montgomerie, Zhewen Yu",
    author_email="am9215@ic.ac.uk",
    description="Streaming Architecture Mapping Optimiser",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AlexMontgomerie/samo",
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        "networkx>=2.5",
        "tqdm>=4.62.3",
        "numpy>=1.19.2",
        "tabulate>=0.8.9",
    ]
)



import csv
import copy
import itertools
from dataclasses import dataclass, field
import numpy as np
from tqdm import tqdm
import time
import os

from samo.model import Network

@dataclass
class RuleBased:
    network: Network

    def update(self):
        for partition in self.network.partitions:
            for index, layer in enumerate(partition):
                partition.nodes[layer]["hw"].update(hw_update=True)

    def optimise_single_partition(self, partition_index):
        print(f"Partition {partition_index}:\n------------\n")

        if not self.network.partitions[partition_index].check_constraints():
            return False

        step = True
        log = []
        while step:
            step = False
            partition = self.network.partitions[partition_index]

            cost = self.network.eval_cost()
            node_latencys = np.array([ partition.nodes[layer]["hw"].latency() for layer in list(partition.nodes())])

            node_index = np.argsort(node_latencys)[-1]
            layer = list(partition.nodes())[node_index]
            node_hw = partition.nodes[layer]["hw"]

            layer_configurations = list(itertools.product(
                node_hw.valid_channel_in_folding,
                node_hw.valid_channel_out_folding,
                node_hw.valid_kernel_folding))

            current_config = [node_hw.channel_in_folding, node_hw.channel_out_folding, node_hw.kernel_folding]
            layer_configurations = list(filter(lambda x: np.prod(x) > np.prod(current_config), layer_configurations))

            if node_hw.constraints["matching_intra_folding"]:
                layer_configurations = list(filter(lambda x: x[0] == x[1], layer_configurations))

            layer_configurations = sorted(layer_configurations, key=lambda x: np.prod(x))

            # uncomment the following code, faster optimiser but worse performance
            #def leq_folding(config):
            #    for i in range(len(config)):
            #        if config[i] < current_config[i]:
            #            return False
            #    return True
            #layer_configurations = list(filter(leq_folding, layer_configurations))

            if len(layer_configurations) > 0:
                step_candidates = {}
                next_folding_candidates = []

                prev_throughput_in = partition.eval_throughput_in()
                prev_throughput_out = partition.eval_throughput_out()
                try_merge_prev = False
                try_merge_next = False

                for config in layer_configurations:
                    # only explore the next and closest folding
                    if len(next_folding_candidates) > 1:
                        break

                    partition = self.network.partitions[partition_index]

                    network_copy = copy.deepcopy(self.network)
                    node_hw = partition.nodes[layer]["hw"]

                    node_hw.channel_in_folding = config[0]
                    partition.folding_match(layer, config[0], "io")
                    node_hw.channel_out_folding = config[1]
                    partition.folding_match(layer, config[1], "io")
                    node_hw.kernel_folding = config[2]

                    # update the network
                    self.update()

                    # check the network is within platform resource constraints
                    if self.network.check_constraints():
                        step_candidates[config] = copy.deepcopy(self.network)
                        next_folding_candidates.append(np.prod(config))
                        next_folding_candidates = list(set(next_folding_candidates))
                    elif not partition.check_memory_bandwdith_constraint():
                        curr_throughput_in = partition.eval_throughput_in()
                        curr_throughput_out = partition.eval_throughput_out()

                        if curr_throughput_in > prev_throughput_in:
                            try_merge_prev = True
                        if curr_throughput_out > prev_throughput_out:
                            try_merge_next = True

                    self.network = network_copy

                step = len(step_candidates) > 0

                # choose the transformation with minimal resource
                chosen = True
                sorted_candidates = dict(sorted(step_candidates.items(), key=lambda kv: kv[1].partitions[partition_index].avg_rsc_util()))
                for config, network in sorted_candidates.items():
                    if chosen:
                        self.network = network

                    # log the current resources and cost
                    new_cost = self.network.eval_cost()
                    #new_resource = self.network.partitions[partition_index].eval_resource()

                    # update the log
                    if chosen:
                        log += [[
                                time.time()-self.start_time,
                                new_cost,
                                #new_resource["BRAM"],
                                #new_resource["DSP"],
                                #new_resource["LUT"],
                                #new_resource["FF"],
                                #chosen,
                                #layer,
                                #config
                        ]]

                    chosen = False
            else:

                try_merge_prev = True
                try_merge_next = True

        partition = self.network.partitions[partition_index]
        if partition.eval_latency()/partition.freq < partition.platform["reconf_time"]:
            partition.try_merge_prev = True
            partition.try_merge_next = True
        else:
            partition.try_merge_prev = try_merge_prev
            partition.try_merge_next = try_merge_next

        # write log to a file
        with open("outputs/log.csv".format(partition_index), "a") as f:

            writer = csv.writer(f)
            [ writer.writerow(row) for row in log ]

        partition.summary()

        return True

    def merge_partitions(self):
        print("resolving memory bound partitions")
        reject_list = []

        while True:
            partitions = copy.deepcopy(self.network.partitions)
            cost = self.network.eval_cost()

            merge_prev_candidates = []
            merge_next_candidates = []

            for partition_index, partition in enumerate(self.network.partitions):
                if partition_index != 0 and partition.try_merge_prev and (partition_index-1, partition_index) not in reject_list:
                    merge_prev_candidates.append(partition_index)
                if partition_index != len(self.network.partitions)-1 and partition.try_merge_next and (partition_index, partition_index+1) not in reject_list:
                    merge_next_candidates.append(partition_index)

            merge_total_candidates = merge_prev_candidates + merge_next_candidates
            merge_total_candidates = list(set(merge_total_candidates))

            if len(merge_total_candidates) == 0:
                break

            partition_latencys = [ self.network.partitions[partition_index].eval_latency() for partition_index in merge_total_candidates]
            partition_index    = merge_total_candidates[partition_latencys.index(max(partition_latencys))]

            if partition_index in merge_next_candidates:
                merge_pair = (partition_index, partition_index+1)
            elif partition_index in merge_prev_candidates:
                merge_pair = (partition_index-1, partition_index)

            print(merge_pair)

            self.network.partitions[merge_pair[0]].reset()
            self.network.partitions[merge_pair[1]].reset()
            self.network.merge(merge_pair)
            status = self.optimise_single_partition(merge_pair[0])

            if not status or self.network.eval_cost() >= cost:
                self.network.partitions = partitions
                reject_list.append(merge_pair)
                print("reject")
            else:
                for i, merge in enumerate(reject_list):
                    if merge[0] >= merge_pair[1]:
                        reject_list[i] = (merge[0]-1,merge[1]-1)
                print("accept")

            print("*******")
    def optimise(self):
        if os.path.exists("outputs/log.csv"):
            os.remove("outputs/log.csv")

        for partition_index in range(len(self.network.partitions)):
            self.optimise_single_partition(partition_index)

        self.merge_partitions()



from dataclasses import dataclass, field
from functools import reduce
from tabulate import tabulate
from typing import List
import networkx as nx
import json
import copy

from networkx.algorithms.dag import ancestors
from networkx.algorithms.dag import descendants

from .partition import Partition

class Network:
    """

    Attributes
    ----------
    reference: Partition
        reference partition to initialse design and validate the
        topology with
    partitions: List[Partition]
        list of partitions that constitute the network
    enable_reconf: Bool
        flag to indicate whether or not to enable reconfiguration
        (more than one partition)
    """

    def __init__(self, reference):
        self.reference = reference
        self.partitions = [copy.deepcopy(reference)]
        self.enable_reconf = True

    def eval_latency(self) -> float:
        """
        latency of the network in microseconds
        """
        return sum([ partition.eval_latency()/partition.freq for partition in self.partitions ]) + \
                sum([ partition.platform["reconf_time"] for partition in self.partitions[:-1] ])

    def eval_throughput(self) -> float:
        """
        throughput of the network in mega-samples per second
        """
        return float(self.batch_size)/self.eval_latency()

    def eval_cost(self) -> float:
        """
        evaluates either throughput or latency based on the
        objective set by the optimisation method
        """
        if self.objective == "throughput":
            return -self.eval_throughput()
        elif self.objective == "latency":
            return self.eval_latency()

    def check_constraints(self):
        """
        checks all the constraints on the network. This includes:

        - all partition constraints
        - partitions are all a subset of the reference network
        """
        # check the partitions make up reference design
        # TODO
        # check constraints of all partitions
        return reduce(lambda a,b: a and b,
                [ partition.check_constraints() for partition in self.partitions ])

    def summary(self):
        """
        prints out a string to the console giving the current state
        of the network in terms of performance and parameters.
        """
        # print the summary for each partition
        for index, partition in enumerate(self.partitions):
            print(f"Partition {index}:\n------------\n")
            partition.summary()
        print(f"Network Summary:\n------------\n")
        print(f"Objective: {self.objective}")
        print(f"Batch Size: {self.batch_size} (img/batch)")
        print(f"Total Latency: {self.eval_latency()} (us/batch)")
        print(f"Total Throughput: {self.eval_throughput()} (img/us)")
        print("\n")

    def split(self, partition_index, nodes):
        """
        method to perform splits on a given partition in the network.
        The `nodes` in the partition chosen by `partition_index` is
        where the split occurs. The `nodes` must be adjacent for the
        split to happen. This creates two seperate partitions in the
        partition list.

        ```
        self.split(0, (node_0, node_1))
        ```
        """
        # get the node indices
        n0 = nodes[0]
        n1 = nodes[1]
        # check nodes are adjacent
        assert self.reference.has_edge(n0, n1)
        # get partition
        p = self.partitions[partition_index]
        # get all nodes before n0 and n1
        p0_nodes = [*list(ancestors(p, n0)), n0]
        p1_nodes = [n1, *list(descendants(p, n1))]
        # get the new partitions
        p0 = p.subgraph(p0_nodes)
        p1 = p.subgraph(p1_nodes)
        # copy other partition information for new partitions
        p0.platform = p.platform
        p0.freq = p.freq
        p0.wordlength = p.wordlength
        p0.constraints = p.constraints
        p1.platform = p.platform
        p1.freq = p.freq
        p1.wordlength = p.wordlength
        p1.constraints = p.constraints
        # check partitions aren't empty
        assert p0.nodes != None
        assert p1.nodes != None
        # remove partition and add new partitions
        self.partitions.remove(p)
        self.partitions.insert(partition_index, p1)
        self.partitions.insert(partition_index, p0)

    def valid_splits(self, partition_index):
        """
        method to return all valid splits on the partition given
        by `partition_index`. This is a list of pairs of adjacent
        nodes.
        """
        if self.enable_reconf:
            return list(self.partitions[partition_index].edges)
        else:
            return []

    def merge(self, partitions):
        """
        merges two partitions where the output node is adjacent to
        the input node of the other. The function excepts a tuple of
        two partition indices.

        ```
        self.merge((0,1))
        ```
        """
        # get the partitions
        p0 = self.partitions[partitions[0]]
        p1 = self.partitions[partitions[1]]
        # check edge between them exists
        assert self.reference.has_edge(p0.output_node, p1.input_node)
        # remove the 2nd partition
        self.partitions.remove(p1)
        # merge the 2nd partition into first
        self.partitions[partitions[0]] = nx.compose(p0,p1)
        self.partitions[partitions[0]].add_edge(p0.output_node, p1.input_node)
        # copy over partition information
        self.partitions[partitions[0]].freq = p0.freq
        self.partitions[partitions[0]].wordlength = p0.wordlength
        self.partitions[partitions[0]].platform = p0.platform
        self.partitions[partitions[0]].constraints = p0.constraints

    def valid_merges(self):
        """
        returns all valid merges that exist in the network as a
        list of tuples of adjacent partition indices.
        """
        # NOTE assuming partitions are sequential
        merges = []
        for i in range(1,len(self.partitions)):
            merges.append((i-1,i))
        return merges

